"""
   DIRAC.Interfaces.API package

   DIRAC - Distributed Infrastructure With Remote Agent Control

   DIRAC is a system which allows to integrate various distributed
   computing resources in a single system which is used for MC data
   production, data reprocessing and final user analysis. It consists
   of the following main subsystems:

     - Workload Management System
     - Data Management System
     - Information Services
     - Tools and utilities

   The DIRAC API provides a transparent and secure way for users to submit
   jobs to the Grid.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"
