#!/usr/bin/env python
# -*- coding:utf-8 -*-
# author: lintao

"""
This module is a helper module for command line interface.

It can help you auto complete the directory name.

The code is a part of

  https://github.com/mirguest/CmdDirectoryCompletion

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
